'From Haver 5.0 [latest update: #4743] on 9 August 2021 at 10:53:29 pm'!
'Description Tests for Klotzli.'!
!provides: 'KlotzliTest' 1 13!
!requires: 'Klotzli' 1 18 nil!
SystemOrganization addCategory: 'KlotzliTest'!


!moduleCreation: KlotzliTest!
Modules newEnvironment: #KlotzliTest!

!interfacesOf: KlotzliTest!
Modules environment: #KlotzliTest ::
	interface: #SPI aliasFor: #API!

!importsOf: KlotzliTest!
Modules environment: #KlotzliTest :: 
	import: #UTI of: Modules>>#Klotzli!

!classDefinition: (Modules>>#KlotzliTest>>#TestEvaluator) category: 'KlotzliTest'!
TestCase subclass: #TestEvaluator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'KlotzliTest'
	inModule: #KlotzliTest!
!classDefinition: (Modules>>#KlotzliTest>>#TestEvaluator) class category: 'KlotzliTest'!
(Modules>>#KlotzliTest>>#TestEvaluator) class
	instanceVariableNames: ''!

!classDefinition: (Modules>>#KlotzliTest>>#TestScopes) category: 'KlotzliTest'!
TestCase subclass: #TestScopes
	instanceVariableNames: 'globalScope undefinedSymbolNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'KlotzliTest'
	inModule: #KlotzliTest!
!classDefinition: (Modules>>#KlotzliTest>>#TestScopes) class category: 'KlotzliTest'!
(Modules>>#KlotzliTest>>#TestScopes) class
	instanceVariableNames: ''!

!classDefinition: (Modules>>#KlotzliTest>>#TestSmalltalkBuilder) category: 'KlotzliTest'!
TestCase subclass: #TestSmalltalkBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'KlotzliTest'
	inModule: #KlotzliTest!
!classDefinition: (Modules>>#KlotzliTest>>#TestSmalltalkBuilder) class category: 'KlotzliTest'!
(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) class
	instanceVariableNames: ''!


!(Modules>>#KlotzliTest>>#TestEvaluator) commentStamp: '<historical>' prior: 0!
Test Klotzli's evaluator.!

!(Modules>>#KlotzliTest>>#TestScopes) commentStamp: '<historical>' prior: 0!
I test the usage of scopes.!

!(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) commentStamp: '<historical>' prior: 0!
Test the SmalltalkBuilder!

!(Modules>>#KlotzliTest>>#TestEvaluator) methodsFor: 'testing' stamp: 'KLG 8/2/2021 20:59:48'!
testLiterals
	"Test the reduction of literals using the standard evaluator"

	#(1 'foo' $A) do: [ :literal |
		self
			assert: literal 
			equals:
				(SmalltalkBuilder source: literal) build 
					reductionResult smalltalkValue ]! !

!(Modules>>#KlotzliTest>>#TestEvaluator) methodsFor: 'testing' stamp: 'KLG 8/9/2021 22:47:31'!
testPrimitives
	"Test the reduction of primitives"

	self
		assert: 42
		equals:
			(SmalltalkBuilder source: #(+ 2 40)) build 
				reductionResult smalltalkValue;
		assert: 'Foo Bar'
		equals:
			(SmalltalkBuilder source: #(#'string-append' 'Foo' ' ' 'Bar')) build 
				reductionResult smalltalkValue! !

!(Modules>>#KlotzliTest>>#TestEvaluator) methodsFor: 'testing' stamp: 'KLG 8/6/2021 20:22:22'!
testVariables
	"Test the reduction of variables using the standard evaluator"

	self
		assert: 42
		equals:
			(SmalltalkBuilder source: #(program (define foo 42) foo)) build 
				reductionResult smalltalkValue! !

!(Modules>>#KlotzliTest>>#TestScopes) methodsFor: 'setUp/tearDown' stamp: 'KLG 8/3/2021 12:47:38'!
setUp
	"Setup a global scope for testting"

	super setUp.
	globalScope _ GlobalScope new! !

!(Modules>>#KlotzliTest>>#TestScopes) methodsFor: 'testing' stamp: 'KLG 8/3/2021 18:49:13'!
testGlobalNames
	"Test the defintion and use of global names."

        | definitionNode symbolNode1 symbolNode2 |
	globalScope defineSymbol: (definitionNode _ DefinitionNode for: #s1).
	self
		assert: (globalScope isIdentifierDefined: #s1);
		assert:  globalScope equals: (globalScope scopeForIdentifier: #s1);
		assert: definitionNode equals: (globalScope definitionOf: #s1);
		assert: nil equals: (globalScope definitionOf: #notThere).
	globalScope addUse: (symbolNode1 _ SymbolNode for: #s1).
	self
		assert: {symbolNode1} asIdentitySet equals: (globalScope usesOf: #s1);
		assert: IdentitySet new equals: (globalScope usesOf: #notThere).
	globalScope addUse: (symbolNode2 _ SymbolNode for: #s1).
	self
		assert: {symbolNode1. symbolNode2} asIdentitySet equals: (globalScope usesOf: #s1);
		assert: IdentitySet new equals: (globalScope usesOf: #notThere).
	globalScope addUse: (undefinedSymbolNode _ SymbolNode for: #undefined).
	self
		assert: {symbolNode1. symbolNode2} asIdentitySet equals: (globalScope usesOf: #s1);
		assert: IdentitySet new equals: (globalScope usesOf: #notThere);
		assert: {undefinedSymbolNode} asIdentitySet equals: (globalScope usesOf: #undefined);
		assert: {undefinedSymbolNode} equals: globalScope undefinedNodes;
		assert: #(undefined) equals: globalScope undefinedSymbols
	! !

!(Modules>>#KlotzliTest>>#TestScopes) methodsFor: 'testing' stamp: 'KLG 8/3/2021 19:01:23'!
testShadowing
	"Test the shadowing of of defintions in sucscopes."

	| subScope parameterNode subSubScope |
	subScope _ globalScope newSubscope.
	self 
		shouldnt: [ globalScope isIdentifierDefined: #s1 ];
		shouldnt: [ subScope isIdentifierDefined: #s1 ].
	subScope addUse: (SymbolNode for: #s1).
	self
		assert: #(s1) equals: subScope undefinedSymbols.
	subScope defineSymbol: (parameterNode _ ParameterNode for: #s1).
	self 
		assert: #() equals: subScope undefinedSymbols;
		shouldnt: [ globalScope isIdentifierDefined: #s1 ];
		should: [ subScope isIdentifierDefined: #s1 ].
         subSubScope _ subScope newSubscope.
	self 
		shouldnt: [ globalScope isIdentifierDefined: #s1 ];
		should: [ subSubScope isIdentifierDefined: #s1 ].
	self
		assert: nil equals: (globalScope definitionOf: #s1);
		assert: parameterNode equals: (subScope definitionOf: #s1);		
		assert: parameterNode equals: (subSubScope definitionOf: #s1);
		assert: nil equals: (globalScope localDefinitionOf: #s1);
		assert: parameterNode equals: (subScope localDefinitionOf: #s1);		
		assert: nil equals: (subSubScope localDefinitionOf: #s1)
! !

!(Modules>>#KlotzliTest>>#TestScopes) methodsFor: 'testing' stamp: 'KLG 8/3/2021 16:31:37'!
testSubScopeCreation
	"Test the creation of sub scopes."

	| subScope |
	self
		assert: ((subScope _ globalScope newSubscope) isKindOf: LocalScope);
		assert: globalScope equals: subScope parent
	! !

!(Modules>>#KlotzliTest>>#TestScopes) methodsFor: 'testing' stamp: 'KLG 8/3/2021 19:11:18'!
testUndefinedInSubScopes
	"The sthe handling of undefined symbols in sub scopes"

	| subScope parameterNode subSubScope |
	subScope _ globalScope newSubscope.
        subSubScope _ subScope newSubscope.
	subScope addUse: (SymbolNode for: #s1).
	subSubScope addUse: (SymbolNode for: #s1).
	self 
		shouldnt: [ globalScope isIdentifierDefined: #s1 ];
		shouldnt: [ subScope isIdentifierDefined: #s1 ];
		shouldnt: [ subSubScope isIdentifierDefined: #s1 ];
		assert: #(s1) equals: subScope undefinedSymbols;
		assert: #(s1) equals: subSubScope undefinedSymbols.
	subScope defineSymbol: (parameterNode _ ParameterNode for: #s1).
	self 
		shouldnt: [ globalScope isIdentifierDefined: #s1 ];
		should: [ subScope isIdentifierDefined: #s1 ];
		should: [ subSubScope isIdentifierDefined: #s1 ];
		assert: #() equals: subScope undefinedSymbols;
		assert: #() equals: subSubScope undefinedSymbols;
		assert: nil equals: (globalScope definitionOf: #s1);
		assert: parameterNode equals: (subScope definitionOf: #s1);
		assert: parameterNode equals: (subSubScope definitionOf: #s1)
! !

!(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) methodsFor: 'testing' stamp: 'KLG 7/30/2021 20:22:03'!
testApplication
	"Test building some simple function applications."

	| nodeToTest  arguments |
	self 
		assert: ((nodeToTest _ SmalltalkBuilder build: #(operator 1 2 3)) isKindOf: ApplicationNode);
		assert: #operator equals: nodeToTest operator identifier;
		assert: nodeToTest == nodeToTest operator parent.
	self assert: (arguments _ nodeToTest arguments) class == Array.
	1 to: 3 do: [ :count | | argument |
		self
			assert: (argument _ arguments at: count) parent == nodeToTest;
			assert: count equals: argument expression smalltalkValue;
			assert: (argument class == ArgumentNode);
			assert: argument callByValue ]! !

!(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) methodsFor: 'testing' stamp: 'KLG 7/30/2021 20:46:05'!
testDefine
	"Test building some define expressions."

	| nodeToTest expression operands |
	self
		assert: ((nodeToTest _ SmalltalkBuilder build: #(define foo 1)) isKindOf: DefinitionNode);
		assert: #foo equals: nodeToTest identifier;
		assert: 1 equals: nodeToTest expression smalltalkValue;
		assert: nodeToTest == nodeToTest expression parent.
	self
		assert: ((nodeToTest _ SmalltalkBuilder build: #(define bar (+ 1 2))) isKindOf: DefinitionNode);
		assert: #bar equals: nodeToTest identifier;
		assert: #+ equals: (expression _ nodeToTest expression) operator identifier;
		assert: (expression isKindOf: ApplicationNode);
		assert: 1 equals: (operands _ expression operands) first smalltalkValue;
		assert: 2 equals: operands second smalltalkValue! !

!(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) methodsFor: 'testing' stamp: 'KLG 7/30/2021 19:39:27'!
testLambda
	"Test building some lambda expressions."

	| nodeToTest statements parameters |
	self 
		assert: ((nodeToTest _ SmalltalkBuilder build: #( (a b) (+ a b))) isKindOf: LambdaNode);
		assert: ((statements _ nodeToTest statements) first isKindOf: ApplicationNode);
		assert: 1 equals: statements size;
		assert: (parameters _ nodeToTest parameters) isArray.
	parameters with: #(a b) do: [ :parameter :name |
		self
			assert: (parameter isKindOf: ParameterNode);
			assert: name equals: parameter identifier;
			assert: parameter callByValue ]! !

!(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) methodsFor: 'testing' stamp: 'KLG 7/28/2021 22:15:27'!
testLiterals
	"Test building some simple literals."

	| nodeToTest |
	self 
		assert: ((nodeToTest _ SmalltalkBuilder build: 1) isKindOf: NumberNode);
		assert: 1 equals: nodeToTest smalltalkValue;
		assert: ((nodeToTest _ SmalltalkBuilder build: 'Sepp') isKindOf: StringNode);
		assert: 'Sepp' equals: nodeToTest smalltalkValue;
		assert: ((nodeToTest _ SmalltalkBuilder build: $C) isKindOf: CharacterNode);
		assert: $C equals: nodeToTest smalltalkValue! !

!(Modules>>#KlotzliTest>>#TestSmalltalkBuilder) methodsFor: 'testing' stamp: 'KLG 7/28/2021 22:37:32'!
testVariables
	"Test building some simple variables."

	| nodeToTest |
	self 
		assert: ((nodeToTest _ SmalltalkBuilder build: #foo) isKindOf: SymbolNode);
		assert: #foo equals: nodeToTest identifier! !
